#include "StdAfx.h"
#include <mmsystem.h>
#include <dxerr9.h>
#include <dsound.h>
#include <crtdbg.h>

#include "DSQuickLib.h"
#include "MP3StreamSound.h"

DWORD WINAPI MP3StreamSound_DoThread(LPVOID pVoid);
#define	SAFE_RELEASE(o)	{if (o){	(o)->Release(); (o) = NULL;	}}
#define	SAFE_DELETE(o)	{if (o){	delete (o); (o) = NULL;	}}
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }
#define SAFE_CLOSEHANDLE(h) { if(h) { CloseHandle(h);   (h)=NULL; } }

CMP3StreamSound::CMP3StreamSound(CDSoundEnv *pEnv, TCHAR *pFilename)
{
	m_pDSoundEnv = pEnv;
	size_t len = _tcslen(pFilename);
	++len;
	m_pFilename = new TCHAR[len];
	_tcscpy_s(m_pFilename,len,pFilename);
	m_hMMIO = NULL;
	m_lStreamStartOffset = 0;
	m_lFileSize = 0;
	m_lFilePos = 0;
	m_bSourceStreamActive = FALSE;

	m_dwDSBufferSize = 0L;
	m_hThread = NULL;
	m_dwThreadID = 0;
	for (int i = 0; i < MP3SOUNDSTREAM_NUM_BLOCKS ; ++i)
		m_hNotificationEvent[i] = NULL;
	m_hTerminator = NULL;
	m_lpDSBuffer = NULL;

	m_dwNumNotification = MP3SOUNDSTREAM_NUM_BLOCKS;	//	obt@̕E쐬r
	m_dwBufferLengthInSec = MP3SOUNDSTREAM_NUM_BLOCKS;		//	ftHg̃obt@PubNPb

	InitializeCriticalSection(&m_CriticalSection);

	m_dwPhase = 0;
	m_bThreadActive = FALSE;
	m_bThreadDone = FALSE;

	m_iSourceBufferSize = 1441;						//	
	m_pMP3Frame = new BYTE[m_iSourceBufferSize];	//	Ki MP3 ł̍őt[TCY

	m_dwSizeConverted = 0;
	m_pPcmBuffer = NULL;
	m_pPcmBufferCurrent = NULL;
	m_dwPcmBufferSizeRemain = 0;
	m_lPCMSizeDetected = -1;

	m_lHistoryRead = 0;
	m_lHistoryWrite = 0;
	m_hAcmStream = NULL;
	ZeroMemory((BYTE*)&m_acmStreamHeader,sizeof(ACMSTREAMHEADER));
	m_bAcmHeaderPrepared = FALSE;

	pEnv->AddSoundObject(this);
}


//  CheckMP3Header
//	MP3 t[wb_͂B
//	MP3 tH[}bgł́AI[fBIf[^́A
//	1152 TvɃt[PʂŃGR[hB
//	et[̐擪ɂ32 rbg̃t[wb_i[
//	ĂB
//	
//	t[wb_̃tH[}bg͈ȉ̂Ƃ
//	Bit :AAAA AAAA AAAB BCCD EEEE FFGH IIJJ KLMM
//  Byte:+---0----+----1----+----2----+----3----+
//  A t[rbg
//  B ID 00:MPEG2.5 01:reserved 10:MPEG2 11:MPEG1
//  C C 11:CT 10:CU 01:CV 00:reserved
//  D یrbg 1:茟ot@0:t
//  E rbg[g EEEe[uQ
//  F TvOgEEEe[uQ
//  G pfBOrbgEEE 1:]ȃXbg܂ 0:܂܂Ȃ
//  H prbgEEEKigp
//  I [h 00:XeI 01:WCgXeI 10:fA`l 11:VO`l
//  J [hgEEEWCgXeI̎
//  K 쌠 1:ی삠 0:یȂ
//  L IWi^Rs[ 1:IWi 0:Rs[
//  M Gt@VX gpׂGt@VX̎
//	
HRESULT	CMP3StreamSound::CheckMP3Header(LPBYTE pID){
	INT iBitRate;
	INT iFS;
	INT	iFrameSize = 0;
	INT	iPadding = 0;
	INT	iChannel = 0;
	INT iVersion;
	static INT iLayer3BitrateTable[][16] = {
		{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0},
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0}
	};
	static INT iFSTable[][3] = {
		{44100, 48000, 32000},
		{22050, 24000, 16000}
	};
	if ((pID[0] != 0xff) || (0xe0 != (pID[1] & 0xe0))){
		_RPT0(_CRT_WARN,_T("MPEGt[wb_sł\n"));
		return E_FAIL;
	}

	//	MPEG o[W̃`FbN
	switch ((pID[1] >> 3) & 0x03) {
	case 3:
		iVersion = 1;
		break;
	case 2:
		iVersion = 2;
		break;
	default:
		_RPT0(_CRT_WARN,_T("MPEGo[Wsł\n"));
		return E_FAIL;
	}

	//	CR݂̂ɑΉ
	if ((pID[1] >> 1 & 0x03) != 1){
		_RPT0(_CRT_WARN,_T("MP3 tH[}bgł͂܂\n"));
		return E_FAIL;
	}
	
	//	I[fBItH[}bg̎擾
	iBitRate = iLayer3BitrateTable[iVersion - 1][pID[2] >> 4 & 15];
	iFS = iFSTable[iVersion - 1][pID[2] >> 2 & 0x03];
	iPadding = pID[2] >> 1 & 0x01;
	iChannel = ((pID[3] >> 6 & 3) == 3) ?(1):(2);

	iFrameSize = (WORD)((1152 * iBitRate * 1000 / iFS) / 8) + iPadding;

	m_iFrameSize = iFrameSize;

	m_wfMP3.wfx.wFormatTag      = WAVE_FORMAT_MPEGLAYER3;
	m_wfMP3.wfx.nChannels       = iChannel;
	m_wfMP3.wfx.nSamplesPerSec  = iFS;
	m_wfMP3.wfx.nAvgBytesPerSec = (iBitRate * 1000) / 8;
	m_wfMP3.wfx.nBlockAlign     = 1;
	m_wfMP3.wfx.wBitsPerSample  = 0;
	m_wfMP3.wfx.cbSize          = MPEGLAYER3_WFX_EXTRA_BYTES;

	m_wfMP3.wID             = MPEGLAYER3_ID_MPEG;
	m_wfMP3.fdwFlags        = iPadding ? MPEGLAYER3_FLAG_PADDING_ON : MPEGLAYER3_FLAG_PADDING_OFF;
	m_wfMP3.nBlockSize      = iFrameSize;
	m_wfMP3.nFramesPerBlock = 1;
	m_wfMP3.nCodecDelay     = 0x571;
	return	S_OK;
}

//
//	MP3t@C̃tH[}bgmFB
//
HRESULT	CMP3StreamSound::CheckMP3Format(){
	if (m_hMMIO == NULL){
		return	E_FAIL;
	}
	BYTE	id3[10];
	m_lFileSize = mmioSeek(m_hMMIO,0,SEEK_END);
	mmioSeek(m_hMMIO,0,SEEK_SET);
	mmioRead(m_hMMIO,(HPSTR)id3,sizeof(id3));
	//	MPEG2 ``FbN
	if (id3[0] == 'I' && id3[1] == 'D' && id3[2] == '3'){
		//	id3v2 `
		LONG lOffset = 10;
		//	tag TCY̎擾EEEBIG ENDIAN
		lOffset += id3[6] << 21;
		lOffset += id3[7] << 14;
		lOffset += id3[8] << 7;
		lOffset += id3[9];
		m_lStreamStartOffset = lOffset;
		m_lFileSize -= lOffset;
	}else{
		//	id3v1 `
		m_lStreamStartOffset = 0;
		mmioSeek(m_hMMIO,128,SEEK_END);
		mmioRead(m_hMMIO,(HPSTR)id3,3);
		if (id3[0] == 'T' && id3[1] == 'A' && id3[2] == 'G'){
			m_lFileSize -= 128;
		}
	}

	//	MP3 tH[}bg̊mF
	mmioSeek(m_hMMIO,m_lStreamStartOffset,SEEK_SET);
	m_lFilePos = m_lStreamStartOffset;
	mmioRead(m_hMMIO,(HPSTR)id3,4);	//	Xg[wb_̓ǂݍ
	CheckMP3Header(id3);
	mmioSeek(m_hMMIO,m_lFilePos,SEEK_SET);
	m_bSourceStreamActive = TRUE;
	return	S_OK;
}

CMP3StreamSound::~CMP3StreamSound(void)
{
	m_bThreadDone = TRUE;
	if (m_hTerminator != NULL){
		SetEvent(m_hTerminator);
		while(m_bThreadActive){
			Sleep(10);
		}
		SAFE_CLOSEHANDLE(m_hThread);
	}
	for (int i = 0; i < MP3SOUNDSTREAM_NUM_BLOCKS ; ++i)
		SAFE_CLOSEHANDLE(m_hNotificationEvent[i]);
	if (m_hAcmStream != NULL){
		if (m_bAcmHeaderPrepared){
			acmStreamUnprepareHeader(m_hAcmStream,&m_acmStreamHeader,0);
			m_bAcmHeaderPrepared = FALSE;
		}
		acmStreamClose(m_hAcmStream,0);
	}

	DeleteCriticalSection(&m_CriticalSection);
	SAFE_CLOSEHANDLE(m_hTerminator);
	SAFE_RELEASE(m_lpDSBuffer);
	SAFE_DELETE_ARRAY(m_pMP3Frame);
	SAFE_DELETE_ARRAY(m_pPcmBuffer);
	if (m_hMMIO != NULL){
		mmioClose(m_hMMIO,0);
		m_hMMIO = NULL;
	}
	SAFE_DELETE_ARRAY(m_pFilename);
	m_pDSoundEnv->RemoveSoundObject(this);
}

//
//	Đ
//
HRESULT	CMP3StreamSound::Prepare(){
	m_hMMIO = mmioOpen(m_pFilename,NULL,MMIO_READ);
	if (m_hMMIO == NULL){
		_RPT0(_CRT_WARN,_T("t@CJ܂ł\n"));
		return	E_FAIL;
	}
	CheckMP3Format();

	//	ĐtH[}bǧ
	m_wfPcm.cbSize = sizeof(WAVEFORMATEX);
	m_wfPcm.wFormatTag = WAVE_FORMAT_PCM;
	m_wfPcm.nSamplesPerSec = m_wfMP3.wfx.nSamplesPerSec;
	if (0 != acmFormatSuggest(NULL, (WAVEFORMATEX*)&m_wfMP3, &m_wfPcm, sizeof(WAVEFORMATEX), ACM_FORMATSUGGESTF_WFORMATTAG|ACM_FORMATSUGGESTF_NSAMPLESPERSEC)){
		goto	e_Fail;
	}

	DWORD	nAlign;
	nAlign = m_wfPcm.nBlockAlign;
	m_dwNotifySize = nAlign * (DWORD)m_wfPcm.nSamplesPerSec;
	m_dwNotifySize *= m_dwBufferLengthInSec;
	m_dwNotifySize /= m_dwNumNotification;
	m_dwNotifySize -= m_dwNotifySize % nAlign;

	for (int i = 0; i < MP3SOUNDSTREAM_NUM_BLOCKS; ++i)
		m_hNotificationEvent[i] =  CreateEvent( NULL, FALSE, FALSE, NULL );

	m_dwSizeConverted = 0;

	if (FAILED(Create())){
		goto	e_Fail;
	}

	m_hTerminator =  CreateEvent( NULL, FALSE, FALSE, NULL );
	m_hThread = CreateThread(NULL,0,::MP3StreamSound_DoThread,(LPVOID)this,0,&m_dwThreadID);

	m_dwPhase = STREAMSOUND_STANDBY;
	return	S_OK;
e_Fail:
	return	E_FAIL;
}



//
//	Xg[obt@̍\z
//
HRESULT	CMP3StreamSound::Create(){
	LPDIRECTSOUND8	pDS;
    DSBUFFERDESC dsbd;
	HRESULT	hr;


    if( m_pDSoundEnv == NULL || m_hMMIO == NULL)
        return CO_E_NOTINITIALIZED;

	m_dwDSBufferSize = m_dwNumNotification * m_dwNotifySize;
    ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize          = sizeof(DSBUFFERDESC);
    dsbd.dwFlags         = DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLPAN |
						   DSBCAPS_CTRLPOSITIONNOTIFY | 
                           DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_LOCSOFTWARE ;
    dsbd.dwBufferBytes   = m_dwDSBufferSize;
    dsbd.guid3DAlgorithm = GUID_NULL;
    dsbd.lpwfxFormat     = &m_wfPcm;
	pDS = m_pDSoundEnv->GetDirectSound();
	hr = pDS->CreateSoundBuffer( &dsbd, &m_lpDSBuffer, NULL );
	SAFE_RELEASE(pDS);
	if (FAILED(hr)){
		_RPT1(_CRT_WARN,_T("CreateSoundBuffer s 0x%x"),hr);
		goto	e_Fail;
	}
	SetNotify();

    m_dwLastPlayPos     = 0;
    m_dwPlayProgress    = 0;
    m_dwNextWriteOffset = 0;
    //m_bFillNextNotificationWithSilence = FALSE;
	FillBufferWithSound(m_lpDSBuffer);

	return	S_OK;
e_Fail:
	return	hr;
}

//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::MP3StreamReadAndDecode()
// Desc: TEht@Cobt@ɓ]B 
//-----------------------------------------------------------------------------
HRESULT	CMP3StreamSound::MP3StreamReadAndDecode(void *buffer,LONG lSize){
	LONG	lWritten = 0;
	HRESULT hr = E_FAIL;
	INT		bits = (m_wfPcm.wBitsPerSample *m_wfPcm.nChannels) / 8;
	BYTE	*pwrite = (BYTE*)buffer;

	if (m_pPcmBuffer == NULL){
		m_dwPcmBufferSize = 1152 * bits;
		m_pPcmBuffer = new BYTE[m_dwPcmBufferSize];
	}
	while (lSize > lWritten){
		//	obt@Ɏc肪Ώ
		if (m_dwPcmBufferSizeRemain > 0){
			LONG	lSizeWrite = lSize - lWritten;
			if (lSizeWrite > m_dwPcmBufferSizeRemain){
				lSizeWrite = m_dwPcmBufferSizeRemain;
			}
			::CopyMemory((BYTE*)pwrite,m_pPcmBufferCurrent,lSizeWrite);
			pwrite += lSizeWrite;
			lWritten += lSizeWrite;
			m_pPcmBufferCurrent += lSizeWrite;
			m_dwPcmBufferSizeRemain -= lSizeWrite;
			if (m_dwPcmBufferSizeRemain <= 0){
				m_pPcmBufferCurrent = NULL;
				m_dwPcmBufferSizeRemain = 0;
			}else{
				continue;
			}
		}
		if (!m_bSourceStreamActive){
			//	ĐIĂc0 Ŗ߂
			LONG lSizeWrite = lSize - lWritten;
			ZeroMemory(pwrite,lSizeWrite);
			lWritten += lSizeWrite;
			break;
		}


		if (m_pPcmBuffer == NULL){
			m_bSourceStreamActive = FALSE;
			continue;
		}
		m_pPcmBufferCurrent = m_pPcmBuffer;
		m_dwPcmBufferSizeRemain = 0;
		BYTE	id[4];
		MMRESULT	mmr;
		mmioRead(m_hMMIO,(HPSTR)id,4);	//	t[wb_ǂݍ
		CheckMP3Header(id);
		if (!m_bAcmHeaderPrepared || m_iFrameSize > m_iSourceBufferSize){
			if (m_bAcmHeaderPrepared){
				acmStreamUnprepareHeader(m_hAcmStream,&m_acmStreamHeader,0);
				acmStreamClose(m_hAcmStream,0);
				m_bAcmHeaderPrepared = FALSE;
			}
			mmr = acmStreamOpen(&m_hAcmStream,NULL,(LPWAVEFORMATEX)&m_wfMP3,&m_wfPcm, NULL, 0, 0, ACM_STREAMOPENF_NONREALTIME);
			if (mmr != 0){
				_RPT1(_CRT_WARN,_T("acmI[vł܂ł G[R[h%d\n"),mmr);
				m_bSourceStreamActive = FALSE;
				continue;
			}

			if (m_pMP3Frame == NULL){
				_RPT0(_CRT_WARN,_T("mۂł܂ł"));
				m_bSourceStreamActive = FALSE;
				continue;
			}
			if (m_iFrameSize > m_iSourceBufferSize){
				SAFE_DELETE_ARRAY(m_pMP3Frame);
				m_pMP3Frame = new BYTE[m_iFrameSize];
				m_iSourceBufferSize = m_iFrameSize;
			}
			
			if (m_pPcmBuffer != NULL){
				ZeroMemory(&m_acmStreamHeader, sizeof(ACMSTREAMHEADER));
				m_acmStreamHeader.cbStruct    = sizeof(ACMSTREAMHEADER);
				m_acmStreamHeader.pbSrc       = m_pMP3Frame;
				m_acmStreamHeader.cbSrcLength = m_iSourceBufferSize;
				m_acmStreamHeader.pbDst       = m_pPcmBuffer;
				m_acmStreamHeader.cbDstLength = m_dwPcmBufferSize;
				acmStreamPrepareHeader(m_hAcmStream,&m_acmStreamHeader,0);
				m_bAcmHeaderPrepared = TRUE;
			}else{
				m_bSourceStreamActive = FALSE;
				continue;
			}
		}
		//	History ̕ۑ
		m_pHistory[m_lHistoryWrite].lFilePos = m_lFilePos;
		m_pHistory[m_lHistoryWrite].lPcmPos = m_dwSizeConverted;
		m_lHistoryWrite = (m_lHistoryWrite + 1) % MP3SOUNDSTREAM_NUM_HISTORY;
		if (m_lHistoryWrite == m_lHistoryRead){
			m_lHistoryRead = (m_lHistoryRead + 1) % MP3SOUNDSTREAM_NUM_HISTORY;
		}
		CopyMemory(m_pMP3Frame,id,4);
		mmioRead(m_hMMIO,(HPSTR)m_pMP3Frame+4,m_iFrameSize - 4);
		DWORD dwLengthSaved = m_acmStreamHeader.cbSrcLength;
		m_acmStreamHeader.cbSrcLength = m_iFrameSize;

		acmStreamConvert(m_hAcmStream, &m_acmStreamHeader, 0);
		m_dwPcmBufferSizeRemain = m_acmStreamHeader.cbDstLengthUsed;
		m_dwSizeConverted += m_acmStreamHeader.cbDstLengthUsed;
		m_acmStreamHeader.cbSrcLength = dwLengthSaved;

		m_lFilePos += m_iFrameSize;
		if ((m_lFilePos - m_lStreamStartOffset) >= m_lFileSize){
			m_bSourceStreamActive = FALSE;
			m_lPCMSizeDetected = m_dwSizeConverted;
		}
	}
	return	S_OK;
}


//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::FillBufferWithSound()
// Desc: TEht@Cobt@ɓ]B 
//-----------------------------------------------------------------------------
HRESULT CMP3StreamSound::FillBufferWithSound( LPDIRECTSOUNDBUFFER pDSB )
{
    HRESULT hr; 
    VOID*   pDSLockedBuffer      = NULL; // bNꂽobt@ւ̃|C^
    DWORD   dwDSLockedBufferSize = 0;    // bNꂽobt@̃TCY

	if( pDSB == NULL )
        return CO_E_NOTINITIALIZED;

    if( FAILED( hr = RestoreBuffer( pDSB, NULL ) ) ) {
        _RPT0(_CRT_WARN,  _T("RestoreBuffer"));
		return	hr;
	}

	//	ĐJn_AĐisJE^ɃRs[
	m_dwPlayProgress = m_dwSizeConverted;

	//	TEhobt@̃bN
	if( FAILED( hr = pDSB->Lock( 0, m_dwDSBufferSize, 
                                 &pDSLockedBuffer, &dwDSLockedBufferSize, 
								 NULL, NULL, 0L ) ) ){
        _RPT1(_CRT_WARN, _T("Lock"), hr );
		return	hr;
	}
	if (m_bAcmHeaderPrepared){
		acmStreamUnprepareHeader(m_hAcmStream,&m_acmStreamHeader,0);
		acmStreamClose(m_hAcmStream,0);
		m_bAcmHeaderPrepared = FALSE;
	}
	MP3StreamReadAndDecode((BYTE*) pDSLockedBuffer,dwDSLockedBufferSize);

    pDSB->Unlock( pDSLockedBuffer, dwDSLockedBufferSize, NULL, 0 );
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::RestoreBuffer()
// Desc: ꂽobt@XgA
//-----------------------------------------------------------------------------
HRESULT CMP3StreamSound::RestoreBuffer( LPDIRECTSOUNDBUFFER pDSB, BOOL* pbWasRestored )
{
    HRESULT hr;

    if( pDSB == NULL )
        return CO_E_NOTINITIALIZED;
    if( pbWasRestored )
        *pbWasRestored = FALSE;

    DWORD dwStatus;
	if( FAILED( hr = pDSB->GetStatus( &dwStatus ) ) ){
        _RPT1(_CRT_WARN,_T("GetStatus %x\n"), hr );
		return	hr;
	}

    if( dwStatus & DSBSTATUS_BUFFERLOST ){
        while(true){
            hr = pDSB->Restore();
            if( hr != DSERR_BUFFERLOST )
				break;
            Sleep( 10 );
        }

        if( pbWasRestored != NULL )
            *pbWasRestored = TRUE;

		m_dwPlayProgress = 0L;

        return S_OK;
    }else{
        return S_FALSE;
    }
}

//
//	CMP3StreamSound::SetNotify()
//	Notification Cxg̎
//
HRESULT	CMP3StreamSound::SetNotify()
{
	HRESULT	hr;
	DSBPOSITIONNOTIFY*  pPosNotify     = NULL; 
	LPDIRECTSOUNDNOTIFY pDSNotify      = NULL;
	DWORD	dwC;
	if( FAILED( hr = m_lpDSBuffer->QueryInterface( IID_IDirectSoundNotify, 
												(VOID**)&pDSNotify ) ) )
	{
		_RPT1(_CRT_WARN,_T("QueryInterface s 0x%x"),hr);
		goto	e_Fail;
	}

	pPosNotify = new DSBPOSITIONNOTIFY[ m_dwNumNotification];
	for( dwC = 0; dwC < m_dwNumNotification; dwC++ )
	{
		pPosNotify[dwC].dwOffset     = (m_dwNotifySize * dwC) + m_dwNotifySize - 1;
		pPosNotify[dwC].hEventNotify = m_hNotificationEvent[dwC];             
	}

	if( FAILED( hr = pDSNotify->SetNotificationPositions( m_dwNumNotification,
														pPosNotify ) ) )
	{
		_RPT0(_CRT_WARN,  _T("SetNotificationPositions") );
		goto	e_Fail;
	}
	SAFE_RELEASE( pDSNotify );
	SAFE_DELETE_ARRAY( pPosNotify );
	return	S_OK;

e_Fail:
	SAFE_RELEASE( pDSNotify );
	SAFE_DELETE_ARRAY( pPosNotify );
	return	hr;
}

//----------------------------------------------------
//	CMP3StreamSound::DoThread()
//	Streaming ̃XbhsI
//----------------------------------------------------
void	CMP3StreamSound::DoThread(){
    DWORD   dwEvent;
	HRESULT	hr;
	int	i;
	m_bThreadActive = TRUE;
	HANDLE	hEvents[MP3SOUNDSTREAM_NUM_BLOCKS + 1];
	for (i = 0; i < MP3SOUNDSTREAM_NUM_BLOCKS ; ++i)
		hEvents[i] = m_hNotificationEvent[i];
	hEvents[i] = m_hTerminator;
	while(!m_bThreadDone){
        dwEvent = WaitForMultipleObjects( MP3SOUNDSTREAM_NUM_BLOCKS + 1, hEvents, FALSE, INFINITE);
		if (dwEvent >= WAIT_OBJECT_0 && dwEvent < WAIT_OBJECT_0 + MP3SOUNDSTREAM_NUM_BLOCKS){
			//_RPT1(_CRT_WARN, _T("Notify:%d\n"),dwEvent - WAIT_OBJECT_0);
			EnterCriticalSection(&m_CriticalSection);
			hr = HandleNotificationEvent();
			LeaveCriticalSection(&m_CriticalSection);
			if (FAILED(hr)){
				m_lpDSBuffer->Stop();
				m_dwPhase = STREAMSOUND_STOP;
			}
		}
		if (dwEvent == WAIT_OBJECT_0 + MP3SOUNDSTREAM_NUM_BLOCKS){
		}
	}
	m_bThreadActive = FALSE;
}

//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::HandleNotificationEvent()
// Desc: Notification ̃Cxgnh
//-----------------------------------------------------------------------------
HRESULT CMP3StreamSound::HandleNotificationEvent(  )
{
    HRESULT hr;
    DWORD   dwCurrentPlayPos;
    DWORD   dwPlayDelta;
    VOID*   pDSLockedBuffer = NULL;
    VOID*   pDSLockedBuffer2 = NULL;
    DWORD   dwDSLockedBufferSize;
    DWORD   dwDSLockedBufferSize2;

    if( m_lpDSBuffer == NULL )
        return CO_E_NOTINITIALIZED;

    // obt@̃XgA
    BOOL bRestored;
	if( FAILED( hr = RestoreBuffer( m_lpDSBuffer, &bRestored ) ) ){
		_RPT0(_CRT_WARN,_T("RestoreBuffer"));
		return	hr;
	}

    if( bRestored )
    {
		if( FAILED( hr = FillBufferWithSound( m_lpDSBuffer ) ) ){
			_RPT0(_CRT_WARN,_T("FillBufferWithSound"));
			return	hr;
		}
        return S_OK;
    }

    // DirectSoundBuffer bN
    if( FAILED( hr = m_lpDSBuffer->Lock( m_dwNextWriteOffset, m_dwNotifySize, 
                                         &pDSLockedBuffer, &dwDSLockedBufferSize, 
										 &pDSLockedBuffer2, &dwDSLockedBufferSize2, 0L ) ) ){
		_RPT0(_CRT_WARN,_T("Lock"));
		return	hr;
	}

	//	m_dwDSBufferSize  m_dwNextWriteOffset ͂ƂɁA
	//	m_dwNotifySize ̔{ł邽߁A pDSLockedBuffer2 ͗LɂȂ鎖͖B
	if( pDSLockedBuffer2 != NULL ){
        return E_UNEXPECTED; 
	}

	if( FAILED( hr = MP3StreamReadAndDecode((void*)pDSLockedBuffer,dwDSLockedBufferSize))){
		m_lpDSBuffer->Unlock( pDSLockedBuffer, dwDSLockedBufferSize, NULL, 0 );
		_RPT0(_CRT_WARN,_T("Read"));
		return	hr;		
	}

    m_lpDSBuffer->Unlock( pDSLockedBuffer, dwDSLockedBufferSize, NULL, 0 );

	if( FAILED( hr = m_lpDSBuffer->GetCurrentPosition( &dwCurrentPlayPos, NULL ) ) ){
		_RPT0(_CRT_WARN,_T("GetCurrentPosition"));
		return	hr;		
	}
	//	O̍Đn_̐isʂ`FbNB
    if( dwCurrentPlayPos < m_dwLastPlayPos )
        dwPlayDelta = ( m_dwDSBufferSize - m_dwLastPlayPos ) + dwCurrentPlayPos;
    else
        dwPlayDelta = dwCurrentPlayPos - m_dwLastPlayPos;

    m_dwPlayProgress += dwPlayDelta;
	//_RPT2(_CRT_WARN,_T("Progress:%x total %x\n"),dwPlayDelta,m_dwPlayProgress);
    m_dwLastPlayPos = dwCurrentPlayPos;

	//	Đn_`FbNA_Ńobt@XgbvB

    if( m_dwPlayProgress >= m_dwSizeConverted/* + m_dwNotifySize*/)
    {
		m_dwPhase = STREAMSOUND_STOP;
        m_lpDSBuffer->Stop();
    }
	//  񃍃bNAhX̍XV
    m_dwNextWriteOffset += dwDSLockedBufferSize; 
    m_dwNextWriteOffset %= m_dwDSBufferSize; // Circular buffer

    return S_OK;
}

//----------------------------------------------------
//	CMP3StreamSound::Play()
//	Streaming ̍ĐJnI
//----------------------------------------------------
HRESULT CMP3StreamSound::Play( DWORD dwPriority, DWORD dwFlags, LONG lVol, LONG lFreq, LONG lPan , LONG lFade){
    HRESULT hr;
    BOOL    bRestored;
    LPDIRECTSOUNDBUFFER pDSB = this->m_lpDSBuffer;
	FLOAT	fFadeTime;
	FLOAT	fVol1, fVol2;
	DWORD	dwCurrent;

	if( pDSB == NULL ){
        _RPT0(_CRT_WARN,_T("No Sound Buffer"));
		return	E_FAIL;
	}
	pDSB->Stop();
	EnterCriticalSection(&m_CriticalSection);
	switch(m_dwPhase){
		case	STREAMSOUND_STANDBY:
		case	STREAMSOUND_STOP:
		case	STREAMSOUND_RUN:
		case	STREAMSOUND_FADEIN:
		case	STREAMSOUND_FADEOUT:
			m_lVol = lVol;
			m_lPan = lPan;
			m_lFreq = lFreq;
			m_dwPhaseAfterReplay = STREAMSOUND_RUN;
			m_fFadeTime = 0.0f;
			m_fFadeLen = (FLOAT)lFade;
			if (lFade != 0){
				m_dwPhaseAfterReplay = STREAMSOUND_FADEIN;
			}
			m_dwPlayProgress = 0;
			if (m_dwPhase != STREAMSOUND_STANDBY){
				Reset();
				FillBufferWithSound(pDSB);
			}
			break;
		case	STREAMSOUND_PAUSE:
			if (!m_bSourceStreamActive && m_dwPlayProgress >= m_dwSizeConverted){
				m_dwPhase = STREAMSOUND_STOP;
				goto	exitPlay;
			}
			dwCurrent = m_dwPlayProgress;
			Reset();
			//	ĐĊJ̃|WVjO
			if (m_lHistoryWrite != m_lHistoryRead){
				INT	i = m_lHistoryWrite;
				INT ix = -1;
				while(i != m_lHistoryRead){
					if (m_pHistory[i].lPcmPos < dwCurrent)
						break;
					ix = i;
					--i;
					if (i < 0)
						i += MP3SOUNDSTREAM_NUM_HISTORY;
				}
				if (ix < 0){
					m_dwPhase = STREAMSOUND_STOP;
					goto	exitPlay;
				}
				m_lFilePos = m_pHistory[ix].lFilePos;
				m_dwSizeConverted = m_pHistory[ix].lPcmPos;
				m_lHistoryRead = m_lHistoryWrite;
				mmioSeek(m_hMMIO,m_lFilePos,SEEK_SET);
				FillBufferWithSound(pDSB);
			}
			break;
		default:
			_RPT0(_CRT_WARN, _T("TEhobt@Đ\ɂȂĂ܂B") );
			hr = E_FAIL;

			goto	exitPlay;
			break;
	}
	pDSB->SetCurrentPosition(0);	

    // Restore the buffer if it was lost
	if( FAILED( hr = RestoreBuffer( pDSB, &bRestored ) ) ){
		_RPT0(_CRT_WARN,_T("RestoreBuffer"));
		goto	exitPlay;
	}

    if( bRestored )
    {
        // The buffer was restored, so we need to fill it with new data
		if( FAILED( hr = FillBufferWithSound( pDSB ) ) ){
			_RPT0(_CRT_WARN,_T("FillBufferWithSound"));
			goto	exitPlay;
		}
    }

	fFadeTime = m_fFadeTime;
	if (fFadeTime > m_fFadeLen)
		fFadeTime = m_fFadeLen;
	switch(m_dwPhaseAfterReplay){
		case	STREAMSOUND_FADEIN:
			fVol1 = DSBVOLUME_MIN * (m_fFadeLen - fFadeTime) / m_fFadeLen;
			fVol2 = (fFadeTime / m_fFadeLen) * m_lVol;
			lVol = (LONG)(fVol1 + fVol2);
			pDSB->SetVolume( lVol );
			break;
		case	STREAMSOUND_FADEOUT:
			fVol1 = DSBVOLUME_MIN * fFadeTime / m_fFadeLen;
			fVol2 = ((m_fFadeLen - fFadeTime) / m_fFadeLen) * m_lVol;
			lVol = (LONG)(fVol1 + fVol2);
			pDSB->SetVolume( lVol );
			break;
		default:
			pDSB->SetVolume( m_lVol );
			break;
	}
    pDSB->SetPan( m_lPan );
    if( m_lFreq != -1 ){
        pDSB->SetFrequency( m_lFreq );
    }
    
	if (SUCCEEDED(hr = pDSB->Play( 0, dwPriority, dwFlags | DSBPLAY_LOOPING ))){
		m_dwPhase = m_dwPhaseAfterReplay;
	}
exitPlay:
	LeaveCriticalSection(&m_CriticalSection);
	return	hr;
}

//----------------------------------------------------
//	CMP3StreamSound::Stop()
//	Streaming ̍ĐII
//----------------------------------------------------
HRESULT CMP3StreamSound::Stop(){
	HRESULT	hr = S_OK;
	switch(m_dwPhase){
		case	STREAMSOUND_RUN:
		case	STREAMSOUND_FADEIN:
		case	STREAMSOUND_FADEOUT:
			EnterCriticalSection(&m_CriticalSection);
            hr = m_lpDSBuffer->Stop();
			if (FAILED(hr)){
				_RPT0(_CRT_WARN,_T("Stop"));
			}
			m_dwPhase = STREAMSOUND_STOP;
			hr = Reset();
			LeaveCriticalSection(&m_CriticalSection);
			break;
		case	STREAMSOUND_PAUSE:
			EnterCriticalSection(&m_CriticalSection);
			m_dwPhase = STREAMSOUND_STOP;
			hr = Reset();
			LeaveCriticalSection(&m_CriticalSection);
			break;
	}
	return	hr;
}

HRESULT	CMP3StreamSound::Rewind(){
	HRESULT	hr;
	if (m_dwPhase == STREAMSOUND_STANDBY)
		return	S_OK;
	hr = Stop();
	if (FAILED(hr))
		return	hr;
	Reset();
	hr = FillBufferWithSound(m_lpDSBuffer);
	return	hr;
}

//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::Reset()
// Desc: obt@̍Đ|WV擪ɖ߂
//-----------------------------------------------------------------------------
HRESULT CMP3StreamSound::Reset()
{
    HRESULT hr;

    if( m_lpDSBuffer == NULL || m_hMMIO == NULL )
        return CO_E_NOTINITIALIZED;

	m_dwLastPlayPos     = 0;
    m_dwPlayProgress    = 0;
    m_dwNextWriteOffset = 0;
    //m_bFillNextNotificationWithSilence = FALSE;

    BOOL bRestored;
	if( FAILED( hr = RestoreBuffer( m_lpDSBuffer, &bRestored ) ) ){
        _RPT0( _CRT_WARN, _T("RestoreBuffer"));
		goto	exitReset;
	}

    //	m_pWaveFile->ResetFile();
	//	UAVI ̍Đ|WV߂ĂI
	//m_dwAviCurrentPos = 0;
	if (m_hMMIO != NULL){
		mmioSeek(m_hMMIO,m_lStreamStartOffset,SEEK_SET);
		m_lFilePos = m_lStreamStartOffset;
		m_bSourceStreamActive = TRUE;
	}

    hr = m_lpDSBuffer->SetCurrentPosition( 0L );  

	m_pPcmBufferCurrent = m_pPcmBuffer;
	m_dwPcmBufferSizeRemain = 0;
exitReset:
	return	hr;
}

//-----------------------------------------------------------------------------
// Name: CStreamSound::FadeOut()
// Desc: FadeOut Jn
//-----------------------------------------------------------------------------
void	CMP3StreamSound::FadeOut(LONG lFadeLen){
	EnterCriticalSection(&m_CriticalSection);
	FLOAT	t;
	switch(m_dwPhase){
		case	STREAMSOUND_RUN:
			m_fFadeTime = 0;
			m_fFadeLen = (FLOAT)lFadeLen;
			m_dwPhase = STREAMSOUND_FADEOUT;
			break;
		case	STREAMSOUND_FADEIN:
			t = m_fFadeTime / m_fFadeLen;
			if (t > 1.0f)
				t = 1.0f;
			m_fFadeLen = (FLOAT)lFadeLen;
			m_fFadeTime = m_fFadeLen - (t * m_fFadeLen);
			m_dwPhase = STREAMSOUND_FADEOUT;
			break;
		case	STREAMSOUND_PAUSE:
			switch(m_dwPhaseAfterReplay){
				case	STREAMSOUND_RUN:
					m_fFadeTime = 0;
					m_fFadeLen = (FLOAT)lFadeLen;
					m_dwPhaseAfterReplay = STREAMSOUND_FADEOUT;
					break;
				case	STREAMSOUND_FADEIN:
					t = m_fFadeTime / m_fFadeLen;
					if (t > 1.0f)
						t = 1.0f;
					m_fFadeLen = (FLOAT)lFadeLen;
					m_fFadeTime = m_fFadeLen - (t * m_fFadeLen);
					m_dwPhaseAfterReplay = STREAMSOUND_FADEOUT;
					break;
				case	STREAMSOUND_FADEOUT:
					break;
			}
			break;
	}
	LeaveCriticalSection(&m_CriticalSection);
}

//-----------------------------------------------------------------------------
// Name: CMP3StreamSound::PollStatus()
// Desc: FADE-OUT, FADE-IN ̏
//-----------------------------------------------------------------------------
void	CMP3StreamSound::PollStatus(FLOAT timeElapsed){
	FLOAT	fFadeTime;
	FLOAT	fVol1, fVol2;
	LONG	lVol;
	switch(m_dwPhase){
		case	STREAMSOUND_FADEIN:
			EnterCriticalSection(&m_CriticalSection);
			if (m_dwPhase == STREAMSOUND_FADEIN){
				fFadeTime = m_fFadeTime + timeElapsed;
				if (fFadeTime > m_fFadeLen){
					m_dwPhase = STREAMSOUND_RUN;
					fFadeTime = m_fFadeLen;
				}
				m_fFadeTime = fFadeTime;
				fVol1 = DSBVOLUME_MIN * (m_fFadeLen - fFadeTime) / m_fFadeLen;
				fVol2 = (fFadeTime / m_fFadeLen) * m_lVol;
				lVol = (LONG)(fVol1 + fVol2);
				m_lpDSBuffer->SetVolume( lVol );
			}
			LeaveCriticalSection(&m_CriticalSection);
			break;
		case	STREAMSOUND_FADEOUT:
			lVol = m_lVol;

			fFadeTime = m_fFadeTime + timeElapsed;
			if (fFadeTime >= m_fFadeLen){
				Stop();
				break;
			}
			m_fFadeTime = fFadeTime;
			EnterCriticalSection(&m_CriticalSection);
			fVol1 = DSBVOLUME_MIN * fFadeTime / m_fFadeLen;
			fVol2 = ((m_fFadeLen - fFadeTime) / m_fFadeLen) * m_lVol;
			lVol = (LONG)(fVol1 + fVol2);
			m_lpDSBuffer->SetVolume( lVol );
			LeaveCriticalSection(&m_CriticalSection);
			if (lVol == DSBVOLUME_MIN){
				Stop();
			}
			break;
	}
}

HRESULT	CMP3StreamSound::Pause(){
    if( m_lpDSBuffer == NULL || m_hMMIO == NULL )
        return CO_E_NOTINITIALIZED;

	EnterCriticalSection(&m_CriticalSection);
	if (m_dwPhase == STREAMSOUND_RUN || m_dwPhase == STREAMSOUND_FADEIN || m_dwPhase == STREAMSOUND_FADEOUT){
		m_lpDSBuffer->Stop();
		m_dwPhaseAfterReplay = m_dwPhase;
		m_dwPhase = STREAMSOUND_PAUSE;
	}
	LeaveCriticalSection(&m_CriticalSection);

	return	S_OK;
}

BOOL	CMP3StreamSound::IsPlaying(){
	return	m_dwPhase == STREAMSOUND_RUN 
		|| m_dwPhase == STREAMSOUND_FADEIN 
		|| m_dwPhase == STREAMSOUND_FADEOUT;
}
DWORD WINAPI MP3StreamSound_DoThread(LPVOID pVoid){
	((CMP3StreamSound*)pVoid)->DoThread();
	return	0L;
}
